/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Map.h"

using namespace FD;
using namespace std;

class MapVectorValueOverwrite;

DECLARE_NODE(MapVectorValueOverwrite)
/*Node
 *
 * @name MapVectorValueOverwrite
 * @category HARK:MISC
 * @description Replaces specific elements of the Map Vector values by a given value.
 *
 * @input_name INPUT
 * @input_type Map<int,ObjectRef>
 * @input_description The input Map. Map<int,Vector<int> >, Map<int,Vector<float> > or Map<int,Vector<complex<float> > >.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The Map with the overwritten value. Map<int,Vector<int> >, Map<int,Vector<float> > or Map<int,Vector<complex<float> > >.
 *
 * @parameter_name OVERWRITTEN_MIN
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index where to start setting the Vector.
 *
 * @parameter_name OVERWRITTEN_MAX
 * @parameter_type int
 * @parameter_value 0
 * @parameter_description The index where to end setting the Vector.
 *
 * @parameter_name OVERWRITE_VALUE_REAL
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Real part of setting value.
 *
 * @parameter_name OVERWRITE_VALUE_IMAG
 * @parameter_type float
 * @parameter_value 0
 * @parameter_description Imaginary part of setting value.
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

class MapVectorValueOverwrite : public BufferedNode {
	int inputID;
	int outputID;
	int vMin;
	int vMax;
	float fReal;
	float fImag;	
	bool debug;

public:
	MapVectorValueOverwrite(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		vMin = dereference_cast<int>(parameters.get("OVERWRITTEN_MIN"));
		vMax = dereference_cast<int>(parameters.get("OVERWRITTEN_MAX"));
		fReal = dereference_cast<float>(parameters.get("OVERWRITE_VALUE_REAL"));
		fImag = dereference_cast<float>(parameters.get("OVERWRITE_VALUE_IMAG"));
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void checkSize(int &vMin, int &vMax, int vSize) {
		/*if (vMin < 0) vMin = 0;
		else if (vMin >= vSize) vMin = vSize - 1;
		if (vMax < 0) vMax = 0;
		else if (vMax >= vSize) vMax = vSize - 1;*/
		if (vMin > vMax) {
			cerr << "******* MapVectorValueOverwrite: Minimun index is larger than maximum index." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MapVectorValueOverwrite]", __FILE__, __LINE__));
		}
		if ((vMin < 0 && vMax < 0) || (vMin >= vSize && vMax >= vSize)) {
			cerr << "******* MapVectorValueOverwrite: Minimun and maximum index are out of range." << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MapVectorValueOverwrite]", __FILE__, __LINE__));
		}
		if (vMin < 0) vMin = 0;
		if (vMax >= vSize) vMax = vSize - 1;
	}

	void calculate(int output_id, int count, Buffer &out) {
		RCPtr<Map<int, ObjectRef> > input_ptr =  getInput(inputID, count);
		const Map<int, ObjectRef> input_map = *input_ptr;

		if (debug == true) {
			cout << "******* MapVectorValueOverwrite debug point-Input check (count=" << count << ")" << endl;
 			if (input_map.empty() == true) {
				cout << "map empty" << endl;
			}
			else {
				cout << "Input map begins from " << (input_map.begin()->first) << endl;
				cout << "Number of channels: " << input_map.size() << endl;
				cout << "Type of first channel of ObjectRef of input map is: ";
				if (typeid(*(input_map.begin()->second))      == typeid(Vector<int>))			  cout << "Vector<int>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Vector<float>))           cout << "Vector<float>" << endl;
				else if (typeid(*(input_map.begin()->second)) == typeid(Vector<complex<float> >)) cout << "Vector<complex<float> >" << endl;
				else                                                                              cout << "unidentified." << endl;
			}
		}

		Map<int, ObjectRef>& output = *new Map<int, ObjectRef>;
		out[count] = &output;

		if (input_map.empty() == false) {
			Map<int, ObjectRef>::const_iterator it = input_map.begin();

			if (typeid(*(it->second)) == typeid(Vector<int>)) {
				for (; it != input_map.end(); ++it) {
					const Vector<int> &vcInput = object_cast<Vector<int> >(it->second);
					size_t vSize = vcInput.size();
					checkSize(vMin, vMax, vSize);
					/*RCPtr<Vector<int> > fbank(Vector<int>::alloc(vSize));*/
					RCPtr<Vector<int> > fbank(new Vector<int>(vSize));
					*fbank = vcInput;
					for (int i = vMin; i <= vMax; i++) {
						(*fbank)[i] = static_cast<int>(fReal);
					} 
					/*int i = 0;
					for (; i < vMin; i++)  (*fbank)[i] = vcInput[i];
					for (; i <= vMax; i++) (*fbank)[i] = static_cast<int>(fReal);
					for (; i < vSize; i++) (*fbank)[i] = vcInput[i];*/
					output[it->first] = fbank;
				}
			}
			else if (typeid(*(it->second)) == typeid(Vector<float>)) {
				for (; it != input_map.end(); ++it) {
					const Vector<float> &vcInput = object_cast<Vector<float> >(it->second);
					size_t vSize = vcInput.size();
					checkSize(vMin, vMax, vSize);
					/*RCPtr<Vector<float> > fbank(Vector<float>::alloc(vSize));*/
					RCPtr<Vector<float> > fbank(new Vector<float>(vSize));
					*fbank = vcInput;
					for (int i = vMin; i <= vMax; i++) {
						(*fbank)[i] = fReal;
					} 
					/*int i = 0;
					for (; i < vMin; i++)  (*fbank)[i] = vcInput[i];
					for (; i <= vMax; i++) (*fbank)[i] = fReal;
					for (; i < vSize; i++) (*fbank)[i] = vcInput[i];*/
					output[it->first] = fbank;
				}
			}
			else if (typeid(*(it->second)) == typeid(Vector<complex<float> >)) {
				for (; it != input_map.end(); ++it) {
					const Vector<complex<float> > &vcInput = object_cast<Vector<complex<float> > >(it->second);
					size_t vSize = vcInput.size();
					checkSize(vMin, vMax, vSize);
					/*RCPtr<Vector<complex<float> > > fbank(Vector<complex<float> >::alloc(vSize));*/
					RCPtr<Vector<complex<float> > > fbank(new Vector<complex<float> >(vSize));
					*fbank = vcInput;
					for (int i = vMin; i <= vMax; i++) {
						(*fbank)[i] = complex<float>(fReal, fImag);
					}
					/*int i = 0;
					for (; i < vMin; i++)  (*fbank)[i] = vcInput[i];
					for (; i <= vMax; i++) (*fbank)[i] = complex<float>(fReal, fImag);
					for (; i < vSize; i++) (*fbank)[i] = vcInput[i];*/
					output[it->first] = fbank;
				}
			}
			else {
				cerr << "******* MapVectorValueOverwrite: Input type is unidentified." << endl;
				throw(new NodeException(NULL, "Exception caught in calculate[MapVectorValueOverwrite]", __FILE__, __LINE__));
			}
		}		

        if (debug == true) cout << "******* MapVectorValueOverwrite debug point-Finish" << endl;
	}

    NO_ORDER_NODE_SPEEDUP(MapVectorValueOverwrite);
};
