/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Map.h"
#include "Source.h"
#include <sstream>
#include <iomanip>

#include <boost/shared_ptr.hpp>

using namespace std;
using namespace FD;
using namespace boost;

class SaveFeatures;

#define ENABLE_DEBUGPRINT
#if defined(ENABLE_DEBUGPRINT)

    #define DEBUG_PRINT(format, args...)                \
    fprintf(stderr, "[%s : %s, line %d] "format"\n",      \
        __FUNCTION__, __FILE__, __LINE__, ##args)

#else

    #define DEBUG_PRINT(format, args...)

#endif // #if defined(ENABLE_DEBUGPRINT)


DECLARE_NODE(SaveFeatures)

#if !defined(ENABLE_HARKIO3)

/*Node
 *
 * @name SaveFeatures
 * @category HARK:FeatureExtraction
 * @description Save a sequence of feature vectors as a binary file per utterace.
 *
 * @input_name FEATURES
 * @input_type Map<int,ObjectRef>
 * @input_description Feature vectors. The key is source ID, the value is a feature vector (Vector<float>).
 *
 * @input_name SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source". Feature vectors are saved for each source ID. This parameter is optional.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The same as FEATURE.
 *
 * @parameter_name BASENAME
 * @parameter_type string
 * @parameter_description Base name for save-file.
 *
END*/

/*
 * @parameter_name OUTPUTTYPE
 * @parameter_type string
 * @parameter_list BINARY
 * @parameter_value BINARY
 * @parameter_description Output type [default: BINARY].
 * 
*/

class SaveFeatures : public BufferedNode {
  int featuresID;
  int outputID;
  int sourcesID;
  string baseName;
  map<int, shared_ptr<ofstream> > streams;
    
public:
  SaveFeatures(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), sourcesID(-1)
  {
    featuresID = addInput("FEATURES");
    outputID = addOutput("OUTPUT");
    baseName = object_cast<String> (parameters.get("BASENAME"));
    inOrder = true;
  }
    
  ~SaveFeatures() {}
    
  void reset()
  {
    streams.clear();
  }

  virtual int translateInput(string inputName)
  {
    for (unsigned int i = 0; i < inputs.size(); i++) {
      if (inputs[i].name == inputName) {
        return i;
      }
    }
    if (inputName == "SOURCES") {
      return sourcesID = addInput(inputName);
    }
    else {
      throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__);
    }
  }
  
  void calculate(int output_id, int count, Buffer &out)
  {
    RCPtr<Map<int, ObjectRef> > input = getInput(featuresID, count);
    RCPtr<Vector<ObjectRef> > sources;
    if (sourcesID != -1) {
      sources = getInput(sourcesID, count);
    }

    if (sources.isNil()) {
      Map<int, ObjectRef>::const_iterator it;
      for (it = input->begin(); it != input->end(); ++it) {
        if (streams.find(it->first) == streams.end()) {
          ostringstream filename;
          filename << baseName << it->first << ".spec";
          streams[it->first].reset(new ofstream(filename.str().c_str()));
        }
        
        ofstream& stream = *streams[it->first];
        
        const Vector<float>& feature = object_cast<Vector<float> >(it->second);
        stream.write((const char*)&feature[0], sizeof(float) * feature.size());
      }
    }
    else {
      for (int i = 0; i < sources->size(); i++) {
        RCPtr<Source> src = (*sources)[i];
        if (streams.find(src->id) == streams.end()) {
          ostringstream filename;
          filename << baseName << src->id << ".spec";
          streams[src->id].reset(new ofstream(filename.str().c_str()));
        }
        ofstream& stream = *streams[src->id];

        Map<int, ObjectRef>::iterator it = input->find(src->id);
        if (it != input->end()) {
          const Vector<float>& feature = object_cast<Vector<float> >(it->second);
          stream.write((const char*)&feature[0], sizeof(float) * feature.size());
        }
      }
    }
    
    if (count % 100 == 0) {
      closeFiles(input);
    }
        
    out[count] = input;
  }

  void closeFiles(RCPtr<Map<int, ObjectRef> > input)
  {
    map<int, shared_ptr<ofstream> >::iterator streams_it = streams.begin();
    while (streams_it != streams.end()) {
      if (input->find(streams_it->first) == input->end()) {
        streams.erase(streams_it++);
      }
      else {
        ++streams_it;
      }
    }
  }
    
  IN_ORDER_NODE_SPEEDUP(SaveFeatures)

};

#else // !defined(ENABLE_HARKIO3)

#include <string.h>
#include "libharkio3.h"

/*Node
 *
 * @name SaveFeatures
 * @category HARK:FeatureExtraction
 * @description Save a sequence of feature vectors as a binary file per utterace.
 *
 * @input_name FEATURES
 * @input_type Map<int,ObjectRef>
 * @input_description Feature vectors. The key is source ID, the value is a feature vector (Vector<float>).
 *
 * @input_name SOURCES
 * @input_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source". Feature vectors are saved for each source ID. This parameter is optional.
 *
 * @output_name OUTPUT
 * @output_type Map<int,ObjectRef>
 * @output_description The same as FEATURE.
 *
 * @parameter_name BASENAME
 * @parameter_type string
 * @parameter_description Base name for save-file.
 *
END*/

/*
 * @parameter_name OUTPUTTYPE
 * @parameter_type string
 * @parameter_list MATRIX:BINARY
 * @parameter_value BINARY
 * @parameter_description Output type [default: BINARY].
 *
*/

class SaveFeatures : public BufferedNode {
  int featuresID;
  int outputID;
  int sourcesID;
  string baseName;
  map<int, harkio_Matrix*> matrixes;
  map<int, char*> matrix_pathes;
  map<int, int> matrix_count;
  map<int, shared_ptr<ofstream> > streams;
  string outputType;
  typedef enum {MATRIX, BINARY} OutputMode;
  OutputMode mode;

public:
  SaveFeatures(string nodeName, ParameterSet params): 
    BufferedNode(nodeName, params),
    sourcesID(-1),
    mode(BINARY)
  {
    featuresID = addInput("FEATURES");
    outputID = addOutput("OUTPUT");
    baseName = object_cast<String> (parameters.get("BASENAME"));
    // issue-HARK-1370 workaround
    //outputType = object_cast<String> (parameters.get("OUTPUTTYPE"));
    outputType = "BINARY";
    if (outputType.compare("MATRIX") == 0) {
      mode = MATRIX;
    } else if (outputType.compare("BINARY") == 0) {
      mode = BINARY;
    }
    inOrder = true;
  }
    
  ~SaveFeatures() {
    map<int, harkio_Matrix*>::iterator it = matrixes.begin();
    map<int, char*>::iterator fit = matrix_pathes.begin();
    while (it != matrixes.end()) {
      char* filename = (*fit).second;
      harkio_Matrix* matrix = (*it).second;
      harkio_Matrix_write(matrix, filename);
      harkio_Matrix_delete(&matrix);
      matrixes.erase(it++);
      matrix_pathes.erase(fit++);
      free(filename);
    }
  }
    
  void reset()
  {
    switch (mode) {
    case MATRIX:
      resetMatrix();
      break;
    case BINARY:
      resetBinary();
      break;
    default:
      break;
    }
  }

  virtual int translateInput(string inputName)
  {
    for (unsigned int i = 0; i < inputs.size(); i++) {
      if (inputs[i].name == inputName) {
        return i;
      }
    }
    if (inputName == "SOURCES") {
      return sourcesID = addInput(inputName);
    }
    else {
      throw new NodeException(this, inputName + " is not supported.", __FILE__, __LINE__);
    }
  }


  void calculate(int output_id, int count, Buffer &out)
  {
    switch (mode) {
    case MATRIX:
      calculateMatrix(output_id, count, out);
      break;
    case BINARY:
      calculateBinary(output_id, count, out);
      break;
    default:
      break;
    }
  }

  void calculateMatrix(int output_id, int count, Buffer &out)
  {
    RCPtr<Map<int, ObjectRef> > input = getInput(featuresID, count);
    RCPtr<Vector<ObjectRef> > sources;
    if (sourcesID != -1) {
      sources = getInput(sourcesID, count);
    }

    if (sources.isNil()) {
      Map<int, ObjectRef>::const_iterator it;
      for (it = input->begin(); it != input->end(); ++it) {
        if (matrixes.find(it->first) == matrixes.end()) {
          ostringstream filename;
          filename << baseName << it->first << ".spec";
          matrix_pathes[it->first] = strdup(filename.str().c_str());
          const Vector<float>& feature = object_cast<Vector<float> >(it->second);          
          matrixes[it->first] = harkio_Matrix_new(strdup("float32"), 1, feature.size());
        }
        
        harkio_Matrix* matrix = matrixes[it->first];
        const Vector<float>& feature = object_cast<Vector<float> >(it->second);
        harkio_Matrix* buf = harkio_Matrix_new(strdup("float32"), 1, feature.size());
        /// @note if possible, needed numpy.array like function.
        for (int i = 0; i < feature.size(); i++) {
          harkio_Matrix_setValueFloat32(buf, 0, i, feature[i]);
        }
        
        matrix = harkio_Matrix_vstack(matrix, buf);
        harkio_Matrix_delete(&buf);
      }
    }
    else {
      for (int i = 0; i < sources->size(); i++) {
        RCPtr<Source> src = (*sources)[i];
        if (matrixes.find(src->id) == matrixes.end()) {
          ostringstream filename;
          filename << baseName << src->id << ".spec";
          matrix_pathes[src->id] = strdup(filename.str().c_str());
          Map<int, ObjectRef>::iterator it = input->find(src->id);
          const Vector<float>& feature = object_cast<Vector<float> >(it->second);          
          matrixes[src->id] = harkio_Matrix_new(strdup("float32"), 1, feature.size());
        }
        harkio_Matrix* matrix = matrixes[src->id];

        Map<int, ObjectRef>::iterator it = input->find(src->id);
        if (it != input->end()) {
          const Vector<float>& feature = object_cast<Vector<float> >(it->second);
          harkio_Matrix* buf = harkio_Matrix_new(strdup("float32"), 1, feature.size());
          for (int i = 0; i < feature.size(); i++) {
            harkio_Matrix_setValueFloat32(buf, 0, i, feature[i]);
          }
          matrix = harkio_Matrix_vstack(matrix, buf);
          harkio_Matrix_delete(&buf);
        }
      }
    }
    
    if (count % 100 == 0) {
      closeMatrixFiles(input);
    }
        
    out[count] = input;
  }
  
  void resetMatrix()
  {
    map<int, harkio_Matrix*>::iterator it = matrixes.begin();
    map<int, char*>::iterator fit = matrix_pathes.begin();
    while (it != matrixes.end()) {
      harkio_Matrix* matrix = (*it++).second;
      char* filename = (*fit++).second;
      harkio_Matrix_delete(&matrix);
      free(filename);
    }
    matrixes.clear();
    matrix_pathes.clear();
  }

  void closeMatrixFiles(RCPtr<Map<int, ObjectRef> > input)
  {
    map<int, harkio_Matrix*>::iterator it = matrixes.begin();
    map<int, char*>::iterator fit = matrix_pathes.begin();
    while (it != matrixes.end()) {
      if (input->find(it->first) != input->end()) {
        char* filename = (*fit).second;
        harkio_Matrix* matrix = (*it).second;
        harkio_Matrix_write(matrix, filename);
        harkio_Matrix_delete(&matrix);
        matrixes.erase(it++);
        matrix_pathes.erase(fit++);
        free(filename);
      } else {
        it++;
        fit++;
      }
    }
  }

  void calculateBinary(int output_id, int count, Buffer &out)
  {
    RCPtr<Map<int, ObjectRef> > input = getInput(featuresID, count);
    RCPtr<Vector<ObjectRef> > sources;
    if (sourcesID != -1) {
      sources = getInput(sourcesID, count);
    }

    if (sources.isNil()) {
      Map<int, ObjectRef>::const_iterator it;
      for (it = input->begin(); it != input->end(); ++it) {
        if (streams.find(it->first) == streams.end()) {
          ostringstream filename;
          filename << baseName << it->first << ".spec";
          streams[it->first].reset(new ofstream(filename.str().c_str()));
        }
        
        ofstream& stream = *streams[it->first];
        
        const Vector<float>& feature = object_cast<Vector<float> >(it->second);
        stream.write((const char*)&feature[0], sizeof(float) * feature.size());
      }
    }
    else {
      for (int i = 0; i < sources->size(); i++) {
        RCPtr<Source> src = (*sources)[i];
        if (streams.find(src->id) == streams.end()) {
          ostringstream filename;
          filename << baseName << src->id << ".spec";
          streams[src->id].reset(new ofstream(filename.str().c_str()));
        }
        ofstream& stream = *streams[src->id];

        Map<int, ObjectRef>::iterator it = input->find(src->id);
        if (it != input->end()) {
          const Vector<float>& feature = object_cast<Vector<float> >(it->second);
          stream.write((const char*)&feature[0], sizeof(float) * feature.size());
        }
      }
    }
    
    if (count % 100 == 0) {
      closeBinaryFiles(input);
    }
        
    out[count] = input;
  }

  void resetBinary()
  {
    streams.clear();
  }

  void closeBinaryFiles(RCPtr<Map<int, ObjectRef> > input)
  {
    map<int, shared_ptr<ofstream> >::iterator streams_it = streams.begin();
    while (streams_it != streams.end()) {
      if (input->find(streams_it->first) == input->end()) {
        streams.erase(streams_it++);
      }
      else {
        ++streams_it;
      }
    }
  }

    
  IN_ORDER_NODE_SPEEDUP(SaveFeatures)

};

#endif // !defined(ENABLE_HARKIO3)
