/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <limits>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "format.h"
#include "Map.h"

using namespace std;
using namespace FD;

class SaveMapFrames;

DECLARE_NODE(SaveMapFrames);
/*Node
 *
 * @name SaveMapFrames
 * @category HARK:MISC
 * @description Save map data of frames to file.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description Data Stream
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The same as input.
 *
 * @parameter_name FILENAME
 * @parameter_type string
 * @parameter_description The file name for saving map data.
 *
 * @parameter_name OUTPUTTYPE
 * @parameter_type string
 * @parameter_list TEXT:RAW
 * @parameter_value TEXT
 * @parameter_description Output file format type [default: TEXT]
 *
 END*/

template <class T>
inline int printLine(const Vector<T> &v, std::ofstream &out) {
    //032318 : This must be used once HARK will be compatible with C++11 
    //out << std::fixed << std::setprecision(std::numeric_limits<float>::max_digits10);
    out << std::fixed << setprecision(9);
    for (size_t i=0; i < v.size(); i++) {
        out << v[i] << " ";
    }
    out << std::endl;
    return 0;
}

template <class T>
inline int printLine(const Vector<complex<T> > &v, std::ofstream &out) {
    //032318 : This must be used once HARK will be compatible with C++11
    //out << std::fixed << std::setprecision(std::numeric_limits<float>::max_digits10);
    out << std::fixed << setprecision(9);
    for (size_t i=0; i < v.size(); i++) {
        out << v[i].real() << " ";
        out << v[i].imag() << " ";

    }
    out << std::endl;
}

class SaveMapFrames : public BufferedNode {
    int inputID;
    int outputID;

    typedef enum {TEXT, RAW} OutputMode;

    OutputMode mode;
    string outputType;
    string filename;
    string formatted_filename;
    bool will_format_filename;
    ofstream fout_in;
    ofstream fout;
    int dimension;
    string containertype;
    vector<string> list;

    public:
    SaveMapFrames(string nodeName, ParameterSet params) : BufferedNode(nodeName, params), mode(TEXT), dimension(1), containertype("Vector")
    {
        inputID = addInput("INPUT");
        outputID = addOutput("OUTPUT");
        outputType = object_cast<String> (parameters.get("OUTPUTTYPE"));
        filename = object_cast<string>(parameters.get("FILENAME"));

        if (filename.c_str() == '\0') {
            throw new NodeException(NULL, string("FILENAME is empty."), __FILE__, __LINE__);
        }

        formatted_filename.clear();
        will_format_filename = false;

        
        if(filename.find("{datatype", 0) != string::npos ||
           filename.find("{colsize", 0) != string::npos ||
           filename.find("{dim", 0) != string::npos ||
           filename.find("{containertype", 0) != string::npos) {
            will_format_filename = true;
        }
        
        if (outputType.compare("RAW") == 0) {
            mode = RAW;
        } else {
            mode = TEXT;
        }
    }

    ~SaveMapFrames()
    {
        fout.close();
    }

    virtual int translateInput(string inputName) {
        if (inputName == "INPUT") {
            for (unsigned int i = 0; i < inputs.size(); i++) {
                if (inputs[i].name == inputName) {
                    return i;
                }
            }
        } else {
            throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);
        }
    }

    // Rosette Lopez : 03132018 Currently not used.  This is for TAG file naming for the main file
    string formatFilename(const char* type, int size) {
        ostringstream new_filename;
        string temp_fname(filename);

        if(will_format_filename) {
            new_filename << fmt::format(temp_fname,
                                              fmt::arg("datatype", type),
                                              fmt::arg("colsize", size),
                                              fmt::arg("dim", dimension),
                                              fmt::arg("containertype", containertype)
                                                ); //format basename
        } else { //use original format if filename is not formatted
            new_filename << temp_fname;
        }
        return new_filename.str();
    }

    string formatFilename(const char* type, int size, int mapKey) {
        ostringstream new_filename;
        ostringstream key;
        string temp_fname(filename);
        key << "_" << setfill('0') << setw(4) << mapKey;

        std::size_t pos = temp_fname.find("_{");
        if (pos != std::string::npos) {
            pos = pos - 1;
            temp_fname.insert(pos, key.str());
        } else {
            temp_fname.append(key.str());
        }

        if(mode == RAW)
            new_filename << temp_fname << "_" << containertype << "_" << type << "_col" << size << ".raw";
        else
            new_filename << temp_fname << "_" << containertype << "_" << type << "_col" << size << ".txt";

        return new_filename.str();
    }

    void handle_files (std::string mname, std::string fname, const int count)
    {

        // List file name into a main/source file
        if (fout.is_open()) {
            fout << fname.c_str() << " ";
        }

        if ( !(std::find(list.begin(), list.end(), fname) != list.end())){
            list.push_back(fname);

            // Open internal file, no append mode since this is the first time the file is opened
            fout_in.open(fname.c_str(), ios::binary);
        }
        else {
            //Open internal file, append new data to existing file
            fout_in.open(fname.c_str(), ios::binary | ios::app);
        }

        if (fout_in.fail()) {
            throw new NodeException(this, string("Can't open file'") + fname + "'.", __FILE__, __LINE__);
        }
    }

    void calculate(int output_id, int count, Buffer &out) {
        ObjectRef input_ptr;
        input_ptr = getInput(inputID, count);
        out[count] = input_ptr;

        // Write into file this iteration frame value
        if (!fout.is_open()) {
            fout.open(filename.c_str(), std::fstream::out);
        }
        if (fout.fail()) {
            throw new NodeException(this, string("Can't open file'") + filename + "'.", __FILE__, __LINE__);
        } else {
            fout << count << " ";
        }
        

        if (typeid(*input_ptr) == typeid(Map<int,ObjectRef>)) {
            RCPtr <Map<int, ObjectRef> > input = getInput(inputID, count);
            for (Map<int, ObjectRef>::const_iterator it = input->begin(); it != input->end(); ++it) {
                string internal_filename("");

                if (typeid(*(it->second)) == typeid(Vector<float>)) {
                    const Vector<float> &tmpvec = object_cast<Vector<float> >(it->second);

                    // Check files if is previously created.  If it is new, store it to filename list for the main file
                    if (!fout_in.is_open()) {
                        internal_filename = formatFilename("float", tmpvec.size(), it->first);
                        handle_files(filename, internal_filename, count);
                    } 

                    // Write Data into file
                    if (mode == RAW) {
                        fout_in.write((const char *) &tmpvec[0], sizeof(float) * tmpvec.size());
                    } else {
                        printLine(tmpvec, fout_in);
                    }
                } else if (typeid(*(it->second)) == typeid(Vector<complex<float> > )) {
                    const Vector<complex<float> > &tmpvec = object_cast<Vector<complex<float> > >(it->second);

                    // Check files if is previously created.  If it is new, store it to filename list for the main file
                    if (!fout_in.is_open()) {
                        internal_filename = formatFilename("complex_float", tmpvec.size(), it->first);
                        handle_files(filename, internal_filename, count);
                    }

                    // Write Data into file
                    if (mode == RAW) {
                        fout_in.write((const char *) &tmpvec[0], sizeof(float) * tmpvec.size() * 2);
                    } else {
                        printLine(tmpvec, fout_in);
                    }
                } else {
                    // Input File Type is not Map <int, ObjectRef>
                    throw new NodeException(this, "Invalid Input Type.", __FILE__, __LINE__);
                }

                // End Internal File, next call might have data stored to another file
                if (fout_in.is_open()) {
                    fout_in.close();
                }
            }
            if (fout.is_open()) { 
                fout << endl;
            }
        } else {
            throw new NodeException(this, "Invalid Input Type", __FILE__, __LINE__);
        }
    }
};

