/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <iomanip>
#include <limits>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Matrix.h"
#include "format.h"

using namespace std;
using namespace FD;

class SaveMatrixFrames;

DECLARE_NODE(SaveMatrixFrames);
/*Node
 *
 * @name SaveMatrixFrames
 * @category HARK:MISC
 * @description Save matrix data of frames to file.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description Data Stream
 *
 * @output_name OUTPUT
 * @output_type any
 * @output_description The same as input.
 *
 * @parameter_name FILENAME
 * @parameter_type string
 * @parameter_description The file name for saving matrix data.
 *
 * @parameter_name OUTPUTTYPE
 * @parameter_type string
 * @parameter_list TEXT:RAW
 * @parameter_value TEXT
 * @parameter_description Output file format type [default: TEXT]
 *
 END*/

template <class T>
inline void printLine(const RCPtr<Matrix<T> > &v, std::ofstream &out) {
    //032318 : This must be used once HARK will be compatible with C++11
    //out << std::fixed << std::setprecision(std::numeric_limits<float>::max_digits10);
    out << std::fixed << setprecision(9);
    for (int row = 0; row < (*v).nrows(); row++) {
        for (int col = 0; col < (*v).ncols(); col++) {
            out << (*v)(row, col) << " ";
        }
    }
    out << std::endl;
}

template <class T>
inline void printLine(const RCPtr<Matrix<complex<T> > > &v, std::ofstream &out) {
    //032318 : This must be used once HARK will be compatible with C++11
    //out << std::fixed << std::setprecision(std::numeric_limits<float>::max_digits10);
    out << std::fixed << setprecision(9);  
    for (int row = 0; row < (*v).nrows(); row++) {
        for (int col = 0; col < (*v).ncols(); col++) {
            out << (*v)(row, col).real() << " ";
            out << (*v)(row, col).imag() << " ";
        }
    }
    out << std::endl;
}

class SaveMatrixFrames : public BufferedNode {
    int inputID;
    int outputID;

    typedef enum {TEXT, RAW} OutputMode;

    OutputMode mode;
    string outputType;
    string filename;
    string formatted_filename;
    bool will_format_filename;
    ofstream fout;
    int dimension;

    public:
    SaveMatrixFrames(string nodeName, ParameterSet params) : BufferedNode(nodeName, params), mode(TEXT), dimension(2)
    {
        inputID = addInput("INPUT");
        outputID = addOutput("OUTPUT");
        outputType = object_cast<String> (parameters.get("OUTPUTTYPE"));
        filename = object_cast<string>(parameters.get("FILENAME"));
        if (filename.c_str() == '\0') {
            throw new NodeException(NULL, string("FILENAME is empty."), __FILE__, __LINE__);
        }

        formatted_filename.clear();
        will_format_filename = false;

        if(filename.find("{datatype", 0) != string::npos ||
           filename.find("{colsize", 0) != string::npos ||
           filename.find("{rowsize", 0) != string::npos ||
           filename.find("{dim", 0) != string::npos) {
            will_format_filename = true;
        }

        if (outputType.compare("RAW") == 0) {
            mode = RAW;
        } else {
            mode = TEXT;
        }
    }

    ~SaveMatrixFrames()
    {
       fout.close();
    }

    virtual int translateInput(string inputName) {
        if (inputName == "INPUT") {
            for (unsigned int i = 0; i < inputs.size(); i++) {
                if (inputs[i].name == inputName) {
                    return i;
                }
            }
        } else {
            throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);
        }
    }

    string formatFilename(const char* type, int row, int col) {
        ostringstream new_filename;

        if(will_format_filename) {
            new_filename << fmt::format(filename,
                                              fmt::arg("datatype", type),
                                              fmt::arg("rowsize", row),
                                              fmt::arg("colsize", col),
                                              fmt::arg("dim", dimension)); //format basename
        } else { //use original format if filename is not formatted
            new_filename << filename;
        }
        return new_filename.str();
    }

    void calculate(int output_id, int count, Buffer &out) {
        ObjectRef input_ptr;
        input_ptr = getInput(inputID, count);
        out[count] = input_ptr;

        if (typeid(*input_ptr) == typeid(Matrix<float>)) {
            RCPtr <Matrix<float> > input = getInput(inputID, count);
            if (!fout.is_open()) {
                formatted_filename = formatFilename("float", (*input).nrows(), (*input).ncols());
                fout.open(formatted_filename.c_str(), ios::binary);
                if (fout.fail()) {
                    throw new NodeException(this, string("Can't open file'") + filename + "'.", __FILE__, __LINE__);
                }
            }

            if (mode == RAW) {
                const Matrix<float> &feature = object_cast<Matrix<float> >(input);
                int row = feature.nrows();
                for (int line = 0; line < row; ++line) {
                    fout.write((const char *) &feature(line, 0), sizeof(float) * feature.ncols());
                }
            } else {
                printLine(input, fout);
            }
        }
        else if (typeid(*input_ptr) == typeid(Matrix<complex<float> >)) {
            RCPtr <Matrix<complex<float> > > input = getInput(inputID, count);
            if (!fout.is_open()) {
                formatted_filename = formatFilename("complex_float", (*input).nrows(), (*input).ncols());
                fout.open(formatted_filename.c_str(), ios::binary);
                if (fout.fail()) {
                    throw new NodeException(this, string("Can't open file'") + filename + "'.", __FILE__, __LINE__);
                }
            }
            if (mode == RAW) {
                const Matrix <complex<float> > &feature = object_cast <Matrix<complex<float> > >(input);
                int row = feature.nrows();
                for (int line = 0; line < row; ++line) {
                    fout.write((const char *) &feature(line, 0), sizeof(float) * feature.ncols() * 2);
                }
            } else {
                printLine(input, fout);
            }
        }
        else {
            throw new NodeException(this, "Invalid Input Type.", __FILE__, __LINE__);
        }
    }
};

