/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <math.h>
#include "BufferedNode.h"
#include "Buffer.h"
#include "Vector.h"
#include "Source.h"

#include <boost/scoped_ptr.hpp>

using namespace std;
using namespace FD;
using namespace boost;

class SourceTransformer;

DECLARE_NODE(SourceTransformer);
/*Node
 *
 * @name SourceTransformer
 * @category HARK:MISC
 * @description Transform source.
 *
 * @input_name INPUT
 * @input_type Vector<ObjectRef>
 * @input_description Source localization results. Each element of the vector is a location without ID specified by "Source".
 *
 * @output_name OUTPUT
 * @output_type Vector<ObjectRef>
 * @input_description Source locations with ID. Each element of the vector is a source location with ID specified by "Source".
 *
 * @parameter_name ENABLE_ID
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Conduct arithmetic operation for id [default: false]
 *
 * @parameter_name ID_OPERATION_TYPE
 * @parameter_type string
 * @parameter_value CONST
 * @parameter_list CONST:ADD:SUB:MUL:DIV
 * @parameter_valid ENABLE_ID=true
 * @parameter_description 
 *
 * @parameter_name ID
 * @parameter_type object
 * @parameter_value <Vector<int> >
 * @parameter_valid ENABLE_ID=true
 * @parameter_description Value for operation. If there is only one element, the operation for all sources of INPUT is done with the element. Alternatively, you can specify the vector so that the operation for each sources of INPUT is done with the different value. Therefore, the number of elements should be one or the same number of elements as INPUT. 
 * 
 * @parameter_name ENABLE_POWER
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Conduct arithmetic operation for power [default: false]
 *
 * @parameter_name POWER_OPERATION_TYPE
 * @parameter_type string
 * @parameter_value CONST
 * @parameter_list CONST:ADD:SUB:MUL:DIV
 * @parameter_valid ENABLE_POWER=true
 * @parameter_description 
 *
 * @parameter_name POWER
 * @parameter_type object
 * @parameter_value <Vector<float> >
 * @parameter_valid ENABLE_POWER=true
 * @parameter_description Value for operation. If there is only one element, the operation for all sources of INPUT is done with the element. Alternatively, you can specify the vector so that the operation for each sources of INPUT is done with the different value. Therefore, the number of elements should be one or the same number of elements as INPUT. 
 *
 * @parameter_name ENABLE_AZIMUTH
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Conduct arithmetic operation for azimuth [default: false]
 *
 * @parameter_name AZIMUTH_OPERATION_TYPE
 * @parameter_type string
 * @parameter_value CONST
 * @parameter_list CONST:ADD:SUB:MUL:DIV
 * @parameter_valid ENABLE_AZIMUTH=true
 * @parameter_description
 *
 * @parameter_name AZIMUTH
 * @parameter_type object
 * @parameter_value <Vector<float> >
 * @parameter_valid ENABLE_AZIMUTH=true
 * @parameter_description Value for operation. If there is only one element, the operation for all sources of INPUT is done with the element. Alternatively, you can specify the vector so that the operation for each sources of INPUT is done with the different value. Therefore, the number of elements should be one or the same number of elements as INPUT. 
 *
 * @parameter_name ENABLE_ELEVATION
 * @parameter_type bool
 * @parameter_value false
 * @parameter_list true:false
 * @parameter_description Conduct arithmetic operation for elevation [default: false]
 *
 * @parameter_name ELEVATION_OPERATION_TYPE
 * @parameter_type string
 * @parameter_value CONST
 * @parameter_list CONST:ADD:SUB:MUL:DIV
 * @parameter_valid ENABLE_ELEVATION=true
 * @parameter_description
 *
 * @parameter_name ELEVATION
 * @parameter_type object
 * @parameter_value <Vector<float> >
 * @parameter_valid ENABLE_ELEVATION=true
 * @parameter_description Value for operation. If there is only one element, the operation for all sources of INPUT is done with the element. Alternatively, you can specify the vector so that the operation for each sources of INPUT is done with the different value. Therefore, the number of elements should be one or the same number of elements as INPUT. 
 *
 END*/

static const int OPERATION_CONST = 0;
static const int OPERATION_ADD = 1;
static const int OPERATION_SUB = 2;
static const int OPERATION_MUL = 3;
static const int OPERATION_DIV = 4;

class SourceTransformer : public BufferedNode {
	int inputID;
	int outputID;

	bool enable_id;
	int operation_id;
	Vector<int> id;

	bool enable_power;
	int operation_power;
	Vector<float> power;

	bool enable_azimuth;
	int operation_azimuth;
	Vector<float> azimuth;

	bool enable_elevation;
	int operation_elevation;
	Vector<float> elevation;

	//bool debug;

public:
	SourceTransformer(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		string type_str;

		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");

		enable_id = dereference_cast<bool>(parameters.get("ENABLE_ID")) ;
		type_str = object_cast<string>(parameters.get("ID_OPERATION_TYPE"));
		if (type_str == "CONST")    operation_id = OPERATION_CONST;
		else if (type_str == "ADD") operation_id = OPERATION_ADD;
		else if (type_str == "SUB") operation_id = OPERATION_SUB;
		else if (type_str == "MUL") operation_id = OPERATION_MUL;
		else if (type_str == "DIV") operation_id = OPERATION_DIV;
		else {
			cerr << "******* SourceTransformer: Input error" << endl;
			throw new NodeException(NULL, "SourceTransformer: ID_OPERATION_TYPE should be \"CONST\" or \"ADD\" or \"SUB\" or \"MUL\" or \"DIV\".", __FILE__, __LINE__);
		}
		id = object_cast<Vector<int> >(parameters.get("ID"));
    
		enable_power = dereference_cast<bool>(parameters.get("ENABLE_POWER")) ;
		type_str = object_cast<string>(parameters.get("POWER_OPERATION_TYPE"));
		if (type_str == "CONST")    operation_power = OPERATION_CONST;
		else if (type_str == "ADD") operation_power = OPERATION_ADD;
		else if (type_str == "SUB") operation_power = OPERATION_SUB;
		else if (type_str == "MUL") operation_power = OPERATION_MUL;
		else if (type_str == "DIV") operation_power = OPERATION_DIV;
		else {
			cerr << "******* SourceTransformer: Input error" << endl;
			throw new NodeException(NULL, "SourceTransformer: POWER_OPERATION_TYPE should be \"CONST\" or \"ADD\" or \"SUB\" or \"MUL\" or \"DIV\".", __FILE__, __LINE__);
		}
		power = object_cast<Vector<float> >(parameters.get("POWER"));
    
		enable_azimuth = dereference_cast<bool>(parameters.get("ENABLE_AZIMUTH")) ;
		type_str = object_cast<string>(parameters.get("AZIMUTH_OPERATION_TYPE"));
		if (type_str == "CONST")    operation_azimuth = OPERATION_CONST;
		else if (type_str == "ADD") operation_azimuth = OPERATION_ADD;
		else if (type_str == "SUB") operation_azimuth = OPERATION_SUB;
		else if (type_str == "MUL") operation_azimuth = OPERATION_MUL;
		else if (type_str == "DIV") operation_azimuth = OPERATION_DIV;
		else {
			cerr << "******* SourceTransformer: Input error" << endl;
			throw new NodeException(NULL, "SourceTransformer: AZIMUTH_OPERATION_TYPE should be \"CONST\" or \"ADD\" or \"SUB\" or \"MUL\" or \"DIV\".", __FILE__, __LINE__);
		}
		azimuth = object_cast<Vector<float> >(parameters.get("AZIMUTH"));
    
		enable_elevation = dereference_cast<bool>(parameters.get("ENABLE_ELEVATION")) ;
		type_str = object_cast<string>(parameters.get("ELEVATION_OPERATION_TYPE"));
		if (type_str == "CONST")    operation_elevation = OPERATION_CONST;
		else if (type_str == "ADD") operation_elevation = OPERATION_ADD;
		else if (type_str == "SUB") operation_elevation = OPERATION_SUB;
		else if (type_str == "MUL") operation_elevation = OPERATION_MUL;
		else if (type_str == "DIV") operation_elevation = OPERATION_DIV;
		else {
			cerr << "******* SourceTransformer: Input error" << endl;
			throw new NodeException(NULL, "SourceTransformer: ELEVATION_OPERATION_TYPE should be \"CONST\" or \"ADD\" or \"SUB\" or \"MUL\" or \"DIV\".", __FILE__, __LINE__);
		}
		elevation = object_cast<Vector<float> >(parameters.get("ELEVATION"));
    
		//debug = dereference_cast<bool>(parameters.get("DEBUG"));

    	inOrder = true;
	}

	void calculate(int output_id, int count, Buffer &out) {
		RCPtr<Vector<ObjectRef> > pcur_src = getInput(inputID, count);
		const Vector<ObjectRef>& cur_sources = *pcur_src;

		//if (debug == true) cout << "******* SourceTransformer debug point-Input check (count=" << count << ")" << endl;

		RCPtr<Vector<ObjectRef> > sources(Vector<ObjectRef>::alloc(0));
		(*(outputs[outputID].buffer))[count] = sources;

		for (int i = 0; i < cur_sources.size(); i++) {
			RCPtr<Source> src = cur_sources[i]->clone();
      
			if (enable_id == true) {
				int i_t = (i >= id.size() ? id.size() - 1 : i );
				int id_i = id[i_t];
				if (operation_id == OPERATION_CONST)    { src->id = id_i; }
				else if (operation_id == OPERATION_ADD) { src->id = src->id + id_i; }
				else if (operation_id == OPERATION_SUB) { src->id = src->id - id_i; }
				else if (operation_id == OPERATION_MUL) { src->id = src->id * id_i; }
				else if (operation_id == OPERATION_DIV) {
					if (id_i == 0) {
						cerr << "******* SourceTransformer: Input error (division by zero)" << endl;
						throw new NodeException(NULL, "SourceTransformer: If ID_OPERATION_TYPEID is \"DIV\", ID should not be zero.", __FILE__, __LINE__);
					} 
					src->id = src->id / id_i;
				}
			}

			if (enable_power == true) {
				float i_t = (i >= power.size() ? power.size() - 1 : i );
				float power_i = power[i_t];
				if (operation_power == OPERATION_CONST)    { src->power = power_i; }
				else if (operation_power == OPERATION_ADD) { src->power = src->power + power_i; }
				else if (operation_power == OPERATION_SUB) { src->power = src->power - power_i; }
				else if (operation_power == OPERATION_MUL) { src->power = src->power * power_i; }
				else if (operation_power == OPERATION_DIV) {
					if (power_i == 0.0f) {
						cerr << "******* SourceTransformer: Input error (division by zero)" << endl;
						throw new NodeException(NULL, "SourceTransformer: If POWER_OPERATION_TYPEID is \"DIV\", POWER should not be zero.", __FILE__, __LINE__);
					} 
					src->power = src->power / power_i;
				}
			}

 			if (enable_azimuth == true) {
				float azimuth_o = atan2(src->x[1], src->x[0]);
				float elevation_o = atan2(src->x[2], sqrt ( src->x[1] * src->x[1] + src->x[0] * src->x[0]));
				float r_o = sqrt(src->x[0] * src->x[0] + src->x[1] * src->x[1] + src->x[2] * src->x[2]);
				float i_t = (i >= azimuth.size()? azimuth.size() - 1 : i);
				float azimuth_i = azimuth[i_t];
				if (operation_azimuth == OPERATION_CONST) {
					src->x[0] = r_o * cos(azimuth_o * M_PI / 180.0f) * cos(elevation_o);
					src->x[1] = r_o * sin(azimuth_o * M_PI / 180.0f) * cos(elevation_o);
					src->x[2] = r_o * sin(elevation_o);
				}
				else if (operation_azimuth == OPERATION_ADD) {
					src->x[0] = r_o * cos(azimuth_o + azimuth_i * M_PI / 180.0f) * cos(elevation_o);
					src->x[1] = r_o * sin(azimuth_o + azimuth_i * M_PI / 180.0f) * cos(elevation_o);
					src->x[2] = r_o * sin(elevation_o);	  
				}
				else if (operation_azimuth == OPERATION_SUB) {
					src->x[0] = r_o * cos(azimuth_o - azimuth_i * M_PI / 180.0f) * cos(elevation_o);
					src->x[1] = r_o * sin(azimuth_o - azimuth_i * M_PI / 180.0f) * cos(elevation_o);
					src->x[2] = r_o * sin(elevation_o);	  
				}
				else if (operation_azimuth == OPERATION_MUL) {
					src->x[0] = r_o * cos(azimuth_o * azimuth_i) * cos(elevation_o);
					src->x[1] = r_o * sin(azimuth_o * azimuth_i) * cos(elevation_o);
					src->x[2] = r_o * sin(elevation_o);	  
				}
				else if (operation_elevation == OPERATION_DIV) {
					if (azimuth_i == 0.0f) {
						cerr << "******* SourceTransformer: Input error (division by zero)" << endl;
						throw new NodeException(NULL, "SourceTransformer: If AZIMUTH_OPERATION_TYPEID is \"DIV\", AZIMUTH should not be zero.", __FILE__, __LINE__);
					} 
					src->x[0] = r_o * cos(azimuth_o / azimuth_i) * cos(elevation_o);
					src->x[1] = r_o * sin(azimuth_o / azimuth_i) * cos(elevation_o);
					src->x[2] = r_o * sin(elevation_o);	  
				}
			}

			if (enable_elevation == true) {
				float azimuth_o = atan2(src->x[1], src->x[0]);
				float elevation_o = atan2(src->x[2], sqrt ( src->x[1] * src->x[1] + src->x[0] * src->x[0]));
				float r_o = sqrt(src->x[0] * src->x[0] + src->x[1] * src->x[1] + src->x[2] * src->x[2]);
				float i_t = (i >= elevation.size()? elevation.size() - 1 : i);
				float elevation_i = elevation[i_t];
				if (operation_elevation == OPERATION_CONST) {
					src->x[0] = r_o * cos(azimuth_o) * cos(elevation_i * M_PI / 180.0f);
					src->x[1] = r_o * sin(azimuth_o) * cos(elevation_i * M_PI / 180.0f);
					src->x[2] = r_o * sin(elevation_i * M_PI / 180.0f);
				}
				else if (operation_elevation == OPERATION_ADD) {
					src->x[0] = r_o * cos(azimuth_o) * cos(elevation_o + elevation_i * M_PI / 180.0f);
					src->x[1] = r_o * sin(azimuth_o) * cos(elevation_o + elevation_i * M_PI / 180.0f);
					src->x[2] = r_o * sin(elevation_o + elevation_i * M_PI / 180.0f);	  
				}
				else if (operation_elevation == OPERATION_SUB) {
					src->x[0] = r_o * cos(azimuth_o) * cos(elevation_o - elevation_i * M_PI / 180.0f);
					src->x[1] = r_o * sin(azimuth_o) * cos(elevation_o - elevation_i * M_PI / 180.0f);
					src->x[2] = r_o * sin(elevation_o - elevation_i * M_PI / 180.0f);	  
				}
				else if (operation_elevation == OPERATION_MUL) {
					src->x[0] = r_o * cos(azimuth_o) * cos(elevation_o * elevation_i);
					src->x[1] = r_o * sin(azimuth_o) * cos(elevation_o * elevation_i);
					src->x[2] = r_o * sin(elevation_o * elevation_i);	  
				}
				else if (operation_elevation == OPERATION_DIV) {
					if (elevation_i == 0.0f) {
						cerr << "******* SourceTransformer: Input error (division by zero)" << endl;
						throw new NodeException(NULL, "SourceTransformer: If ELEVATION_OPERATION_TYPEID is \"DIV\", ELEVATION should not be zero.", __FILE__, __LINE__);
					} 
					src->x[0] = r_o * cos(azimuth_o) * cos(elevation_o / elevation_i);
					src->x[1] = r_o * sin(azimuth_o) * cos(elevation_o / elevation_i);
					src->x[2] = r_o * sin(elevation_o / elevation_i);	  
				}
			}

			/*if (debug) {
				cout << "<Source " << endl;
				cout << "<x " << src->x[0] << " y " << src->x[1] << " z " << src->x[2] << ">" << endl;
				cout << "<power " << src->power << ">" << endl;
				cout << "<id " << src->id << ">" << endl;
				cout << ">" << endl;
			}*/

			sources->push_back(src);
		}

        //if (debug == true) cout << "******* SourceTransformer debug point-Finish" << endl;
	}

	IN_ORDER_NODE_SPEEDUP(SourceTransformer);
};
